from logging import ERROR, WARNING

from django.conf import settings
from django.db import models
from django.db.models.base import ModelBase
from django.utils import timezone

from locking.exceptions import AlreadyLocked
from locking.models import NonBlockingLock

from .managers import TriggerManager
from .exceptions import ProcessLaterError
from .loggers import get_logger
from .loggers.base import TriggerLogger


class TriggerBase(ModelBase):
    """
    A meta class for all Triggers. Adds a default manager that filters
    on type.
    """
    def __new__(cls, name, bases, attrs):
        super_new = super(TriggerBase, cls).__new__
        typed = attrs.pop('typed', None)
        if typed is not None:
            attrs['objects'] = TriggerManager(typed)
        new_class = super_new(cls, name, bases, attrs)
        if typed is None:
            return new_class

        new_class.typed = typed
        return new_class


class Trigger(models.Model):
    """
    A persistent Trigger that needs processing.

    Triggers are created when a certain state is reached internally
    or externally. Triggers are persistent and need processing after
    a certain point in time.

    To create a trigger start by subclassing Trigger and setting the
    'typed' attribute. 'typed' should be a unique slug that identifies
    the trigger. Subclasses should also implement '_process()'.

    Subclasses can be proxy models when no extra data needs to be
    stored. Otherwise use regular subclassing. This will create an
    additional table with trigger specific data and a one-to-one
    relation to the triggers table.

    'source' is a free-form field that can be used to uniquely determine
    the source of the trigger.

    There is a logging framework included. If passed a logger argument in the
    __init__, that logger will be used, otherwise the class' _logger_class
    determines the logger (if any). Default is no logging.
    """
    __metaclass__ = TriggerBase

    # Set typed in a subclass to make it a typed trigger.
    typed = None

    trigger_type = models.CharField(max_length=50, db_index=True)
    source = models.CharField(max_length=150, null=True, blank=True, db_index=True)
    date_received = models.DateTimeField(default=timezone.now)
    date_processed = models.DateTimeField(null=True, blank=True, db_index=True)
    process_after = models.DateTimeField(null=True, blank=True, db_index=True)
    number_of_tries = models.IntegerField(default=0)
    successful = models.NullBooleanField(default=None)

    _logger_class = None

    def __init__(self, *args, **kwargs):
        super(Trigger, self).__init__(*args, **kwargs)
        if hasattr(self, 'typed'):
            self.trigger_type = self.typed

        if self._logger_class:
            self.logger = get_logger(self._logger_class)
        else:
            self.logger = TriggerLogger()

    def set_source(self, *args):
        args = [str(arg) for arg in args]
        self.source = '$'.join(args)

    def get_source(self):
        return tuple(x for x in self.source.split('$') if x != '')

    def process(self, force=False, logger=None, dictionary=None, use_statsd=False):
        """
        Executes the Trigger
        :param bool force: force the execution
        :param string logger: slug of preferred logger
        :param dict dictionary: dictionary needed by trigger to execute
        :param bool use_statsd: whether to use statsd
        :return: None
        """
        dictionary = {} if dictionary is None else {}

        # The task gets locked because multiple tasks in the queue can process the same trigger.
        # The lock assures no two tasks can process a trigger simultaneously.
        # The check for date_processed assures a trigger is not executed multiple times.
        try:
            with NonBlockingLock.objects.acquire_lock(self):
                if logger:
                    self.logger = get_logger(logger)
                now = timezone.now()
                if not force and self.date_processed is not None:
                    # trigger has already been processed. So everything is fine
                    return
                if not force and self.process_after and self.process_after >= now:
                    raise ProcessLaterError(self.process_after)

                try:
                    # execute trigger
                    self.logger.log_result(self, self._process(dictionary))
                    self._handle_execution_success(use_statsd)
                except ProcessLaterError as e:
                    self.process_after = e.process_after
                    self.save()
                except Exception as e:
                    self._handle_execution_failure(e, use_statsd)
                    raise
        except AlreadyLocked:
            pass

    def _process(self, dictionary):
        raise NotImplementedError()

    def _handle_execution_failure(self, exception, use_statsd=False):
        """
        Handle execution failure of the trigger
        :param Exception exception: the exception raised during failure
        :param bool use_statsd: whether to use statsd
        :return: None
        """
        self.number_of_tries += 1
        # Log message if starts retrying too much
        if self.number_of_tries > getattr(settings, 'DJTRIGGERS_TRIES_BEFORE_WARNING', 3):
            # Set a limit for retries
            if self.number_of_tries >= getattr(settings, 'DJTRIGGERS_TRIES_BEFORE_ERROR', 5):
                # Set date_processed so it doesn't retry anymore
                self.date_processed = timezone.now()
                self.successful = False
                level = ERROR
            else:
                level = WARNING

            message = 'Processing of {trigger_type} {trigger_key} raised a {exception_type} (try nr. {try_count})'.\
                format(trigger_type=self.trigger_type, trigger_key=self.pk, exception_type=type(exception).__name__,
                       try_count=self.number_of_tries)
            self.logger.log_message(self, message, level=level)

        # Send stats to statsd if necessary
        if use_statsd:
            from django_statsd.clients import statsd
            statsd.incr('triggers.{trigger_type}.failed'.format(trigger_type=self.trigger_type))

        self.save()

    def _handle_execution_success(self, use_statsd=False):
        """
        Handle execution success of the trigger
        :param bool use_statsd: whether to use statsd
        :return: None
        """
        if self.date_processed is None:
            now = timezone.now()
            self.date_processed = now

        # Send stats to statsd if necessary
        if use_statsd:
            from django_statsd.clients import statsd
            statsd.incr('triggers.{trigger_type}.processed'.format(trigger_type=self.trigger_type))
            if self.date_processed and self.process_after:
                statsd.timing('triggers.{trigger_type}.process_delay_seconds'.format(trigger_type=self.trigger_type),
                              (self.date_processed - self.process_after).total_seconds())

        self.successful = True
        self.save()

    def __repr__(self):
        return 'Trigger {trigger_id} of type {trigger_type} ({is_processed}processed)'.format(
            trigger_id=self.id, trigger_type=self.trigger_type, is_processed='' if self.date_processed else 'not ')


class TriggerResult(models.Model):
    trigger = models.ForeignKey(Trigger, on_delete=models.CASCADE)
    result = models.TextField()

    def __repr__(self):
        return self.result



from django.utils import timezone
from factory import DjangoModelFactory

from djtriggers.tests.models import DummyTrigger


class DummyTriggerFactory(DjangoModelFactory):
    class Meta:
        model = DummyTrigger

    trigger_type = 'dummy_trigger_test'
    source = 'tests'
    date_received = timezone.now()
    date_processed = None
    process_after = None
    number_of_tries = 0



from django.db import models


class TriggerManager(models.Manager):
    def __init__(self, trigger_type):
        super(TriggerManager, self).__init__()
        self.trigger_type = trigger_type

    def get_queryset(self):
        qs = super(TriggerManager, self).get_queryset()
        if self.trigger_type:
            qs = qs.filter(trigger_type=self.trigger_type)
        return qs

    def get_unprocessed_triggers(self):
        qs = self.get_queryset()
        return qs.filter(date_processed__isnull=True)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import datetime


class Migration(migrations.Migration):

    dependencies = [
        ('djtriggers', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='trigger',
            name='date_received',
            field=models.DateTimeField(default=datetime.datetime.now),
            preserve_default=True,
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.5 on 2017-10-03 09:51
from __future__ import unicode_literals

import datetime
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    replaces = [('djtriggers', '0001_initial'), ('djtriggers', '0002_auto_20151208_1454'), ('djtriggers', '0003_auto_20160512_0847'), ('djtriggers', '0004_auto_20170216_1007'), ('djtriggers', '0005_trigger_successful'), ('djtriggers', '0006_auto_20171003_0945')]

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Trigger',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('trigger_type', models.CharField(db_index=True, max_length=50)),
                ('source', models.CharField(blank=True, max_length=150, null=True)),
                ('date_received', models.DateTimeField()),
                ('date_processed', models.DateTimeField(blank=True, db_index=True, null=True)),
                ('process_after', models.DateTimeField(blank=True, db_index=True, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='TriggerResult',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('result', models.TextField()),
                ('trigger', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='djtriggers.Trigger')),
            ],
        ),
        migrations.AlterField(
            model_name='trigger',
            name='date_received',
            field=models.DateTimeField(default=datetime.datetime.now),
        ),
        migrations.AlterField(
            model_name='trigger',
            name='source',
            field=models.CharField(blank=True, db_index=True, max_length=150, null=True),
        ),
        migrations.AddField(
            model_name='trigger',
            name='number_of_tries',
            field=models.IntegerField(default=0),
        ),
        migrations.AlterField(
            model_name='trigger',
            name='date_received',
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
        migrations.AddField(
            model_name='trigger',
            name='successful',
            field=models.NullBooleanField(default=None),
        ),
        migrations.AlterField(
            model_name='trigger',
            name='source',
            field=models.CharField(blank=True, db_index=True, max_length=150, null=True),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djtriggers', '0002_auto_20151208_1454'),
    ]

    operations = [
        migrations.AlterField(
            model_name='trigger',
            name='source',
            field=models.CharField(db_index=True, max_length=250, null=True, blank=True),
        ),
    ]



from logging import log, info

from djtriggers.loggers.base import TriggerLogger


class DatabaseLogger(TriggerLogger):
    def log_result(self, trigger, message, level=None):
        from djtriggers.models import TriggerResult
        TriggerResult.objects.create(trigger=trigger, result=message)

    def log_message(self, trigger, message, level=None):
        if level:
            log(level, message)
        else:
            info(message)


def _prettify(results):
    try:
        return '\n'.join([str(r) for r in results])
    except TypeError:
        return str(results)


class DatabaseSerializeLogger(DatabaseLogger):
    def log_result(self, trigger, results, level=None):
        if results is None:
            return

        from djtriggers.models import TriggerResult
        TriggerResult.objects.create(trigger=trigger, result=_prettify(results))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Trigger',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('trigger_type', models.CharField(max_length=50, db_index=True)),
                ('source', models.CharField(max_length=250, null=True, blank=True)),
                ('date_received', models.DateTimeField()),
                ('date_processed', models.DateTimeField(db_index=True, null=True, blank=True)),
                ('process_after', models.DateTimeField(db_index=True, null=True, blank=True)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='TriggerResult',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('result', models.TextField()),
                ('trigger', models.ForeignKey(to='djtriggers.Trigger')),
            ],
            options={
            },
            bases=(models.Model,),
        ),
    ]



from datetime import timedelta
from mock import patch

from django.test import override_settings
from django.test.testcases import TestCase
from django.utils import timezone

from djtriggers.logic import process_triggers
from djtriggers.tests.factories.triggers import DummyTriggerFactory


class SynchronousExecutionTest(TestCase):
    def setUp(self):
        self.now = timezone.now()

    def test_process_after_now(self):
        trigger = DummyTriggerFactory()
        process_triggers()
        trigger.refresh_from_db()
        assert trigger.date_processed is not None

    def test_process_after_yesterday(self):
        trigger = DummyTriggerFactory(process_after=self.now - timedelta(days=1))
        process_triggers()
        trigger.refresh_from_db()
        assert trigger.date_processed is not None

    def test_process_after_tomorrow(self):
        trigger = DummyTriggerFactory(process_after=self.now + timedelta(days=1))
        process_triggers()
        trigger.refresh_from_db()
        assert trigger.date_processed is None

    def test_already_processed(self):
        trigger = DummyTriggerFactory(date_processed=self.now)
        process_triggers()
        trigger.refresh_from_db()
        assert trigger.date_processed == self.now


class AsynchronousExecutionTest(TestCase):
    def setUp(self):
        self.now = timezone.now()

    @override_settings(DJTRIGGERS_ASYNC_HANDLING=True)
    def test_process_after_now(self):
        trigger = DummyTriggerFactory()
        with patch('djtriggers.logic.process_trigger.apply_async') as process_trigger_patch:
            process_triggers()
            process_trigger_patch.assert_called_once_with((trigger.id, 'djtriggers', 'DummyTrigger'),
                                                          {'use_statsd': False}, max_retries=0)

    @override_settings(DJTRIGGERS_ASYNC_HANDLING=True)
    def test_process_after_yesterday(self):
        trigger = DummyTriggerFactory(process_after=self.now - timedelta(days=1))
        with patch('djtriggers.logic.process_trigger.apply_async') as process_trigger_patch:
            process_triggers()
            process_trigger_patch.assert_called_once_with((trigger.id, 'djtriggers', 'DummyTrigger'),
                                                          {'use_statsd': False}, max_retries=0)

    @override_settings(DJTRIGGERS_ASYNC_HANDLING=True)
    def test_process_after_tomorrow(self):
        DummyTriggerFactory(process_after=self.now + timedelta(days=1))
        with patch('djtriggers.logic.process_trigger.apply_async') as process_trigger_patch:
            process_triggers()
            assert not process_trigger_patch.called

    @override_settings(DJTRIGGERS_ASYNC_HANDLING=True)
    def test_already_processed(self):
        DummyTriggerFactory(date_processed=self.now)
        with patch('djtriggers.logic.process_trigger.apply_async') as process_trigger_patch:
            process_triggers()
            assert not process_trigger_patch.called



from celery import shared_task
from celery.utils.log import get_task_logger

from django.apps import apps


logger = get_task_logger(__name__)


@shared_task
def process_triggers():
    from .logic import process_triggers as process
    process()


@shared_task
def clean_triggers():
    from .logic import clean_triggers as clean
    clean()


@shared_task
def process_trigger(trigger_id, trigger_app_label, trigger_class, *args, **kwargs):
    apps.get_model(trigger_app_label, trigger_class).objects.get(id=trigger_id).process(*args, **kwargs)



class AlreadyProcessedError(Exception):
    pass


class ProcessError(Exception):
    pass


class ProcessLaterError(ProcessError):
    def __init__(self, process_after, *args):
        super(ProcessLaterError, self).__init__(*args)
        self.process_after = process_after



from django.conf import settings
from django.core.urlresolvers import get_mod_func

REGISTRY = {}

loggers = getattr(settings, 'DJTRIGGERS_LOGGERS', ())

for entry in loggers:
    module_name, class_name = get_mod_func(entry)
    logger_class = getattr(__import__(module_name, {}, {}, ['']), class_name)
    instance = logger_class()
    REGISTRY[class_name] = instance


def get_logger(slug):
    return REGISTRY.get(slug, None)



class TriggerLogger(object):
    """
    A logger for the processing of a trigger.
    """

    def log_result(self, trigger, message, level=None):
        """
        Log the result of a trigger.
        :param level:
        """
        pass

    def log_message(self, trigger, message, level=None):
        """
        Log any message during the processing of a trigger.
        """
        pass



from datetime import timedelta
from logging import ERROR, WARNING

from mock import patch
from pytest import raises
from django.test import override_settings
from django.test.testcases import TestCase
from django.utils import timezone
from locking.models import NonBlockingLock

from djtriggers.exceptions import ProcessLaterError
from djtriggers.loggers.base import TriggerLogger
from djtriggers.models import Trigger
from djtriggers.tests.factories.triggers import DummyTriggerFactory


class TriggerTest(TestCase):
    def test_handle_execution_success(self):
        trigger = DummyTriggerFactory()
        trigger._handle_execution_success()

        assert trigger.date_processed
        assert trigger.successful is True

    @patch('django_statsd.clients.statsd')
    def test_handle_execution_success_use_statsd(self, mock_statsd):
        trigger = DummyTriggerFactory(process_after=timezone.now())
        trigger._handle_execution_success(use_statsd=True)

        assert trigger.date_processed
        assert trigger.successful is True
        mock_statsd.incr.assert_called_with(
            'triggers.{trigger_type}.processed'.format(trigger_type=trigger.trigger_type))
        mock_statsd.timing.assert_called_with(
            'triggers.{trigger_type}.process_delay_seconds'.format(trigger_type=trigger.trigger_type),
            (trigger.date_processed - trigger.process_after).total_seconds())

    def test_handle_execution_failure(self):
        trigger = DummyTriggerFactory()
        original_tries = trigger.number_of_tries
        trigger._handle_execution_failure(Exception())

        assert trigger.number_of_tries == original_tries + 1

    @override_settings(DJTRIGGERS_TRIES_BEFORE_WARNING=3)
    @patch.object(TriggerLogger, 'log_message')
    def test_handle_execution_failure_dont_raise_when_under_max_retries(self, mock_logger):
        exception = Exception()
        trigger = DummyTriggerFactory(number_of_tries=2)
        original_tries = trigger.number_of_tries
        trigger._handle_execution_failure(exception)

        assert trigger.number_of_tries == original_tries + 1
        assert not mock_logger.called

    @override_settings(DJTRIGGERS_TRIES_BEFORE_WARNING=3)
    @patch.object(TriggerLogger, 'log_message')
    def test_handle_execution_failure_tries_exceeded(self, mock_logger):
        exception = Exception()
        trigger = DummyTriggerFactory(number_of_tries=3)
        original_tries = trigger.number_of_tries
        trigger._handle_execution_failure(exception)

        assert trigger.number_of_tries == original_tries + 1
        message = ('Processing of {trigger_type} {trigger_key} '
                   'raised a {exception_type} (try nr. {try_count})').format(trigger_type=trigger.trigger_type,
                                                                             trigger_key=trigger.pk,
                                                                             exception_type=type(exception).__name__,
                                                                             try_count=original_tries+1)
        mock_logger.assert_called_with(trigger, message, level=WARNING)

    @override_settings(DJTRIGGERS_TRIES_BEFORE_WARNING=3)
    @override_settings(DJTRIGGERS_TRIES_BEFORE_ERROR=5)
    def test_handle_execution_failure_tries_limit(self):
        """
        When trigger reaches retries limit and ERROR is raised check message level and trigger successful state.
        """
        exception = Exception()
        exception_name = type(exception).__name__
        trigger = DummyTriggerFactory(number_of_tries=3)
        original_tries = trigger.number_of_tries

        with patch.object(TriggerLogger, 'log_message') as mock_logger:
            trigger._handle_execution_failure(exception)
            exceeded_retries = original_tries + 1
            assert trigger.number_of_tries == exceeded_retries
            assert trigger.successful is None
            message = ('Processing of {trigger_type} {trigger_key} '
                       'raised a {exception_type} (try nr. {try_count})').format(trigger_type=trigger.trigger_type,
                                                                                 trigger_key=trigger.pk,
                                                                                 exception_type=exception_name,
                                                                                 try_count=exceeded_retries)
            mock_logger.assert_called_once_with(trigger, message, level=WARNING)

        # do an extra retry
        with patch.object(TriggerLogger, 'log_message') as mock_logger:
            trigger._handle_execution_failure(exception)
            exceeded_retries = original_tries + 2
            assert trigger.number_of_tries == exceeded_retries
            assert trigger.successful is False
            message = ('Processing of {trigger_type} {trigger_key} '
                       'raised a {exception_type} (try nr. {try_count})').format(trigger_type=trigger.trigger_type,
                                                                                 trigger_key=trigger.pk,
                                                                                 exception_type=exception_name,
                                                                                 try_count=exceeded_retries)
            mock_logger.assert_called_once_with(trigger, message, level=ERROR)

    def test_handle_execution_failure_above_tries_limit(self):
        """
        Check if ERROR is still raised for a trigger above retry limit
        """
        exception = Exception()
        exception_name = type(exception).__name__
        trigger = DummyTriggerFactory(number_of_tries=5)
        original_tries = trigger.number_of_tries
        exceeded_retries = original_tries + 1

        with self.settings(DJTRIGGERS_TRIES_BEFORE_WARNING=3) and self.settings(DJTRIGGERS_TRIES_BEFORE_ERROR=5) and \
                patch.object(TriggerLogger, 'log_message') as mock_logger:
            trigger._handle_execution_failure(exception)
            assert trigger.number_of_tries == exceeded_retries
            assert trigger.successful is False
            message = ('Processing of {trigger_type} {trigger_key} '
                       'raised a {exception_type} (try nr. {try_count})').format(trigger_type=trigger.trigger_type,
                                                                                 trigger_key=trigger.pk,
                                                                                 exception_type=exception_name,
                                                                                 try_count=exceeded_retries)
            mock_logger.assert_called_once_with(trigger, message, level=ERROR)

    @patch('django_statsd.clients.statsd')
    def test_handle_execution_failure_use_statsd(self, mock_statsd):
        exception = Exception()
        trigger = DummyTriggerFactory()
        original_tries = trigger.number_of_tries
        trigger._handle_execution_failure(exception, use_statsd=True)

        assert trigger.number_of_tries == original_tries + 1
        assert trigger.successful is None
        mock_statsd.incr.assert_called_with('triggers.{trigger_type}.failed'.format(trigger_type=trigger.trigger_type))

    @patch.object(TriggerLogger, 'log_result')
    def test_process(self, mock_logger):
        trigger = DummyTriggerFactory()
        trigger.process()

        mock_logger.assert_called_with(trigger, trigger._process({}))

    @patch.object(TriggerLogger, 'log_result')
    def test_process_already_processed(self, mock_logger):
        '''Reprocessing already processed triggers should just do nothing.'''
        trigger = DummyTriggerFactory(date_processed=timezone.now())
        assert trigger.date_processed is not None
        assert not mock_logger.called

    def test_process_process_later(self):
        trigger = DummyTriggerFactory(process_after=timezone.now() + timedelta(minutes=1))
        with raises(ProcessLaterError):
            trigger.process()

    @patch.object(Trigger, '_handle_execution_failure')
    def test_process_exception_during_execution(self, mock_fail):
        trigger = DummyTriggerFactory()
        with patch.object(trigger, '_process', side_effect=Exception), raises(Exception):
            trigger.process()
        assert mock_fail.called

    @patch.object(TriggerLogger, 'log_result')
    def test_process_locked(self, mock_logger):
        trigger = DummyTriggerFactory()
        with NonBlockingLock.objects.acquire_lock(trigger):
            trigger.process()

        assert not mock_logger.called



# -*- coding: utf-8 -*-
# Generated by Django 1.9.13 on 2017-08-24 08:48
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djtriggers', '0004_auto_20170216_1007'),
    ]

    operations = [
        migrations.AddField(
            model_name='trigger',
            name='successful',
            field=models.NullBooleanField(default=None),
        ),
    ]



from logging import getLogger

from django.core.management.base import NoArgsCommand

from djtriggers.logic import clean_triggers


logger = getLogger(__name__)


class Command(NoArgsCommand):
    help = clean_triggers.__doc__.strip()

    def handle_noargs(self, **options):
        nr_deleted = clean_triggers()
        logger.info('Cleaned up %s expired triggers', nr_deleted)



from logging import getLogger

logger = getLogger(__name__)


def run_checks(checks):
    """
    Run a number of checks.

    :param tuple checks: a tuple of tuples, with check name and parameters dict.

    :returns: whether all checks succeeded, and the results of each check
    :rtype: tuple of (bool, dict)
    """
    results = {}
    all_succeeded = True
    for check, kwargs in checks:
        succeeded = check(**kwargs).succeeded
        if isinstance(succeeded, bool):
            if not succeeded:
                all_succeeded = False
            results[check.__name__] = succeeded
        elif isinstance(succeeded, tuple):
            if not succeeded[0]:
                all_succeeded = False
            results[check.__name__] = succeeded
    return (all_succeeded, results)


class Check(object):
    def run(self):
        self._result = self._run()
        return self._result

    def _run(self):
        raise NotImplementedError()

    @property
    def result(self):
        if not self.has_run:
            return self.run()
        return self._result

    @property
    def has_run(self):
        return hasattr(self, '_result')

    @property
    def succeeded(self):
        return bool(self.result)

    def __nonzero__(self):
        return self.succeeded



from django.core.management.base import NoArgsCommand

from optparse import make_option

from djtriggers import logic


class Command(NoArgsCommand):
    option_list = NoArgsCommand.option_list + (
        make_option('--use-statsd', dest='use_statsd', action='store_true', default=False,
                    help='Send stats about processing to Statsd'),
    )

    def handle_noargs(self, **options):
        """
        Process all triggers in order of trigger type. This blocks while
        processing the triggers.
        """
        logic.process_triggers(use_statsd=options['use_statsd'])



# -*- coding: utf-8 -*-
# Generated by Django 1.9.7 on 2017-02-16 10:07
from __future__ import unicode_literals

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('djtriggers', '0003_auto_20160512_0847'),
    ]

    operations = [
        migrations.AddField(
            model_name='trigger',
            name='number_of_tries',
            field=models.IntegerField(default=0),
        ),
        migrations.AlterField(
            model_name='trigger',
            name='date_received',
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
    ]



from inspect import isabstract
from logging import getLogger

from dateutil.relativedelta import relativedelta

from django.apps import apps
from django.conf import settings
from django.db import connections
from django.db.models import Q
from django.utils import timezone

from .models import Trigger
from .exceptions import ProcessError, ProcessLaterError
from .tasks import process_trigger


logger = getLogger(__name__)


def process_triggers(use_statsd=False, function_logger=None):
    """
    Process all triggers that are ready for processing.

    :param bool use_statsd: whether to use_statsd
    :return: None
    """
    process_async = getattr(settings, 'DJTRIGGERS_ASYNC_HANDLING', False)

    # Get all triggers that need to be processed
    for model in apps.get_models():
        # Check whether it's a trigger
        if not issubclass(model, Trigger) or getattr(model, 'typed', None) is None or isabstract(model):
            continue

        # Get all triggers of this type that need to be processed
        triggers = model.objects.filter(Q(process_after__isnull=True) | Q(process_after__lt=timezone.now()),
                                        date_processed__isnull=True)

        # Process each trigger
        for trigger in triggers:
            try:
                # Process the trigger, either synchronously or in a Celery task
                if process_async:
                    process_trigger.apply_async((trigger.id, trigger._meta.app_label, trigger.__class__.__name__),
                                                {'use_statsd': use_statsd},
                                                max_retries=getattr(settings, 'DJTRIGGERS_CELERY_TASK_MAX_RETRIES', 0))
                else:
                    trigger.process()

                # Send stats to statsd if necessary
                if use_statsd:
                    from django_statsd.clients import statsd
                    statsd.incr('triggers.{}.processed'.format(trigger.trigger_type))
                    if trigger.date_processed and trigger.process_after:
                        statsd.timing('triggers.{}.process_delay_seconds'.format(trigger.trigger_type),
                                      (trigger.date_processed - trigger.process_after).total_seconds())
            # The trigger didn't need processing yet
            except ProcessLaterError:
                pass
            # The trigger raised an (expected) error while processing
            except ProcessError:
                pass


def clean_triggers(expiration_dt=None, type_to_table=None):
    """
    Clean old processed triggers from the database.

    Args:
        expiration_dt (optional datetime): triggers processed before this timestamp will be cleaned up.
            Defaults to 2 months before the current time.
        type_to_table (optional dict): maps trigger type to database table name.
            Defaults to DJTRIGGERS_TYPE_TO_TABLE django setting.

    `type_to_table` contains has information about which trigger has information
    in which table. This setting is a dict with the trigger types as keys and two
    options for values:

    - a string containing the table where the trigger information is stored
      (with a trigger_ptr_id to link it)
    - a tuple containing elements of two possible types:
        - a string containing the table where the trigger information is stored
          (with a trigger_ptr_id to link it)
        - a tuple containing a tablename and an id field

    Example:

    {'simple_trigger': 'simple_trigger_table',
     'complex_trigger': ('complex_trigger_table1', 'complex_trigger_table2'),
     'complexer_trigger': (('complexer_trigger_table1', 'complexer_trigger_id'), 'complexer_trigger_table2'),
    }
    """
    if expiration_dt is None:
        expiration_dt = timezone.now() - relativedelta(months=2)

    if type_to_table is None:
        type_to_table = getattr(settings, 'DJTRIGGERS_TYPE_TO_TABLE', {})

    cursor = connections['default'].cursor()
    sentinel = object()
    nr_deleted = 0

    for trigger in Trigger.objects.filter(date_processed__lt=expiration_dt):
        # Delete custom trigger information
        table = type_to_table.get(trigger.trigger_type, sentinel)
        if isinstance(table, tuple):
            for t in table:
                if isinstance(t, tuple):
                    cursor.execute('DELETE FROM {} WHERE {} = {}'.format(t[0], t[1], trigger.id))
                else:
                    cursor.execute('DELETE FROM {} WHERE trigger_ptr_id = {}'.format(t, trigger.id))
        elif table != sentinel:
            cursor.execute('DELETE FROM {} WHERE trigger_ptr_id = {}'.format(table, trigger.id))

        # Delete the trigger from the main table
        trigger.delete()
        nr_deleted += 1

    return nr_deleted



from djtriggers.loggers.base import TriggerLogger

from logging import log, info


class PythonLogger(TriggerLogger):
    """
    Logger using the default python logger.
    """
    def log_result(self, trigger, message, level=None):
        if level:
            log(level, message)
        else:
            info(message)

    def log_message(self, trigger, message, level=None):
        if level:
            log(level, message)
        else:
            info(message)



from setuptools import setup, find_packages
from pip.req import parse_requirements
from pip.download import PipSession

from os import path
from djtriggers import __version__


# Lists of requirements and dependency links which are needed during runtime, testing and setup
install_requires = []
tests_require = []
dependency_links = []

# Inject test requirements from requirements_test.txt into setup.py
requirements_file = parse_requirements(path.join('requirements', 'requirements.txt'), session=PipSession())
for req in requirements_file:
    install_requires.append(str(req.req))
    if req.link:
        dependency_links.append(str(req.link))

# Inject test requirements from requirements_test.txt into setup.py
requirements_test_file = parse_requirements(path.join('requirements', 'requirements_test.txt'), session=PipSession())
for req in requirements_test_file:
    tests_require.append(str(req.req))
    if req.link:
        dependency_links.append(str(req.link))


setup(
    name='django-triggers',
    version=__version__,
    url='https://github.com/vikingco/django-triggers',
    license='BSD',
    description='Framework to create and process triggers.',
    long_description=open('README.md', 'r').read(),
    author='Unleashed NV',
    author_email='operations@unleashed.be',
    packages=find_packages('.'),
    include_package_data=True,
    zip_safe=False,  # Don't create egg files, Django cannot find templates in egg files.
    install_requires=install_requires,
    setup_requires=['pytest-runner', ],
    tests_require=tests_require,
    dependency_links=dependency_links,
    classifiers=[
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Operating System :: OS Independent',
        'Environment :: Web Environment',
        'Framework :: Django',
    ],
)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.5 on 2017-10-03 09:45
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('djtriggers', '0005_trigger_successful'),
    ]

    operations = [
        migrations.AlterField(
            model_name='trigger',
            name='source',
            field=models.CharField(blank=True, db_index=True, max_length=150, null=True),
        ),
    ]
